#!/usr/bin/python3
import threading
import sys
import time
import os
import datetime
import zipfile
import subprocess
import logging
import signal
import getopt
import re
import shutil
import ctypes
import configparser

def swmCurSite(loc):
  try:
    cfg = configparser.ConfigParser(strict=False)  # Ignore potential duplicate keys in conf
    cfg.read(loc)
    return cfg.get("global", "site")
  except:
    return "generic"
  
# Handler for the Interrupt signal with required parameters
def keyboard_interrupt(signum, frame):
  log(logger, 'Caught signal {} (shutdown)'.format(signum))
  global interrupted
  interrupted = True
  
def logLifeEvents(site_name, life_event):
  # if i had more time, i would write a shorter letter
  found_site = False
  #found_life_event = False
  if not os.path.exists("log/sys_diagnostics.xml"): 
    if not os.path.exists("log/"):
      os.mkdir("log")    
    diagnostics_tag = etree.Element("diagnostics")
  else:
    try: 
      diagnostics_tag = etree.parse("log/sys_diagnostics.xml").getroot()
    except:
      pass
    diag_site_tags = diagnostics_tag.findall("site")
    for diag_site_tag in diag_site_tags:
      if diag_site_tag.attrib["name"] == site_name:
        try:
          diag_site_tag.find(life_event).text = str(int(diag_site_tag.find(life_event).text) + 1)
        except:
          diag_life_event_tag = etree.Element(life_event)
          diag_life_event_tag.text = "1"
          diag_site_tag.append(diag_life_event_tag)
        #found_life_event = True
        found_site = True
        break
  if not found_site:
    diag_site_tag = etree.Element("site")
    diag_site_tag.attrib["name"] = site_name
    diag_life_event_tag = etree.Element(life_event)
    diag_life_event_tag.text = "1"
    diag_site_tag.append(diag_life_event_tag)
    diagnostics_tag.append(diag_site_tag)
  fd = open("log/sys_diagnostics.xml", "w")
  fd.write(etree.tostring(diagnostics_tag, pretty_print=True).decode())
  fd.close()
  
def parse_module(module):
  typ = ""
  try:
    typ = module.attrib["type"]
  except:
    #print("A module with an invalid (or no) name or type.")
    return None
  params = dict()
  module_children = module.getchildren()
  for module_child in module_children:
    value = None
    param_name = module_child.tag
    #first, let's get "the special cases" out of the way
    if typ == "rcv" and param_name == "offset_from_boss":
      value = dict()
      try:
        offsets_from_boss = module_child.findall("lane")
        for lane in offsets_from_boss:
          mp = 1
          try:
            mp = int(lane.attrib["mp"])
          except:
            pass
          if lane.attrib["num"] not in value:
            value[lane.attrib["num"]] = dict()
          value[lane.attrib["num"]][mp] = float(lane.text)
      except:
        continue
      if len(value) == 0:
        # consistency
        value = None
    #############################################    
    elif typ == "pic" and (param_name == "default_cam_types" or param_name == "default_cam_names"):
      value = list()
      cam_infos = module_child.getchildren()
      for cam_info in cam_infos:
        value.append(cam_info.text)
      if len(value) == 0:
        # consistency
        value = None
    #############################################        
    elif typ == "out" and param_name == "filter_rules":
      # might aswell keep the default
      try:
        elements = module.find("filter_rules").findall("element")
        for element in elements:
          if value == None:
            value = list()
          if "is_sibling_of" in element.attrib:
            value.append((element.text, "is_sibling_of", element.attrib["is_sibling_of"]))
          elif "not_sibling_of" in element.attrib:
            value.append((element.text, "not_sibling_of", element.attrib["not_sibling_of"]))
          else:
            value.append((element.text,))
      except:
        continue
    #############################################    
    elif typ == "cam" and param_name == "ignore_classes":
      class_lane_dict = dict()
      for lane in module_child.getchildren():
        classes = list()
        for cls in lane.getchildren():
          try:
            classes.append(int(cls.text))
          except:
            pass
        class_lane_dict[int(lane.attrib["num"])] = classes
      if len(class_lane_dict) != 0:
        value = class_lane_dict
    #############################################    
    elif typ == "cam" and param_name == "fixed_offset":
      num_lane_dict = dict()
      for lane in module_child.getchildren():
        num_lane_dict[int(lane.attrib["num"])] = int(lane.text)
      if len(num_lane_dict) != 0:
        value = num_lane_dict
    #############################################    
    elif typ == "cam" and param_name == "cam_focus":
      num_lane_dict = dict()
      for lane in module_child.getchildren():
        mp = 1
        try:
          mp = int(lane.attrib["mp"])
        except:
          pass
        if int(lane.attrib["num"]) not in num_lane_dict:
          num_lane_dict[int(lane.attrib["num"])] = dict()
        num_lane_dict[int(lane.attrib["num"])][mp] = float(lane.text)
      if len(num_lane_dict) != 0:
        value = num_lane_dict
    #############################################    
    elif typ == "cam" and param_name == "lanes":
      lanes = list()
      for lane in module_child.getchildren():
        lanes.append(int(lane.text))
      if len(lanes) != 0:
        value = lanes
    #############################################   
    elif typ == "lrm" and param_name == "timeouts":
      timeouts = dict()
      timeout_nodes = list()
      try:
        timeout_nodes = module.find("timeouts").findall("timeout")
      except:
        pass
      impl_timeout_nodes = list()
      try:
        impl_timeout_nodes = module.find("timeouts").findall("impl_interval_timeout")
      except:
        pass
      if len(timeout_nodes) == len(impl_timeout_nodes):
        pairs = list(zip(timeout_nodes, impl_timeout_nodes))
      for pair in pairs:
        timeouts[pair[0].attrib["lane"]] = (pair[0].text, pair[1].text)
      if len(timeouts) != 0:
        value = timeouts
    ############################################# 
    elif typ == "can" and param_name == "board_ids":
      board_ids = list()
      try:
        board_id_nodes = module.find("board_ids").findall("board_id")
        for node in board_id_nodes:
          board_ids.append(int(node.text))
      except:
        pass
      if len(board_ids) != 0:
        value = board_ids
    ############################################# 
    elif param_name == "downstream_modules":
      # handle this elswhere (for now)
      continue
    ############################################# 
    else:
      try:
        value = int(module_child.text)
      except:
        try:
          value = float(module_child.text)
        except:
          value = module_child.text
    if value != None:
      params[param_name] = value
  return params

# Disable QuickEdit mode by setting it's flag to 0x00
kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)
kernel32.SetConsoleMode(kernel32.GetStdHandle(-10), (0x4 | 0x80 | 0x20 | 0x2 | 0x10 | 0x1 | 0x00 | 0x100))
del kernel32  # We don't need this variable floating around anymore, so we delete it

# Get the script parameters
try:
  opts, args = getopt.getopt(sys.argv[1:], 'hc:dmu', ['help', 'conf=', 'debug', 'memory', 'update'])
except getopt.GetoptError:
  print('Parsing of parameters failed. Running in default mode.')
  opts = list()

i_conf_path = ""

# update libraries
LOG_LEVEL = logging.INFO
print_memory_usage = False
for opt, arg in opts:
  if opt in ('-h', '--help'):
    print('Options:\n'
          '-d --debug: Run I with debug output.\n'
          '-c --conf: Specify conf path (bypass default).\n'
          '-h --help: Print help and exit.\n'
          '-m --memory: Uses guppy to print basic memory information once per minute. guppy must be installed manually!'
          '-u --update: Install libraries in requirements-{}.txt and update backup.py to newest version.'.format(sys.version_info[0]))
    sys.exit()
  elif opt in ('-d', '--debug'):
    LOG_LEVEL = logging.DEBUG
  elif opt in ('-c', '--conf'):
    i_conf_path = arg.replace("\\", "/")
  elif opt in ('-m', '--memory'):
    print_memory_usage = True
  elif opt in ('-u', '--update'):
    libraries_updated = False
    if sys.version_info[0] == 2:
      os.environ["PYTHONIOENCODING"] = "utf-8"
    try:
      subprocess.check_call('{} -m pip install --upgrade pip'.format(sys.executable))  # We have to set default Python encoding for this to work
    except (subprocess.CalledProcessError, LookupError):
      os.execv(sys.executable, [sys.executable] + sys.argv)
    try:  # Update siwim_backup
      PACKAGE_NAME = 'siwim_backup'
      SCRIPT_FOLDER = 'D:/siwim_mkiii/backup_script'
      SCRIPT = os.path.join(SCRIPT_FOLDER, 'backup.py')

      output = subprocess.check_output('{} -m pip download siwim-backup'.format(sys.executable)).decode()
      archive = re.findall(r'{}-\d\.\d.\d\.zip'.format(PACKAGE_NAME), output)[0]

      file_path = '{}/{}/backup_win.py'.format(archive.rsplit('.', 1)[0], PACKAGE_NAME)
      try:
        os.makedirs(SCRIPT_FOLDER)
      except:  # We don't care if the file exists.
        pass
      with zipfile.ZipFile(archive, 'r') as zip:
        zip.extract(file_path)

      shutil.move(file_path, SCRIPT)
      os.remove(archive)
      shutil.rmtree(archive.rsplit('.', 1)[0])
    except Exception as e:
      print('Downloading siwim_backup failed: {}'.format(e))
    while not libraries_updated:  # In most cases this will only be run once, but we can't let the application run if updating libraries fails.
      try:
        subprocess.check_call([sys.executable, '-m', 'pip', 'install', '-r', 'requirements-{}.txt'.format(sys.version_info[0])])
        libraries_updated = True
      except Exception as e:  # If this fails for whatever reason just keep on going ...
        print('Updating of libraries failed: {}\nRetrying ...'.format(e))
    os.execv(sys.executable, [sys.executable] + [f for f in sys.argv if f not in ['-u', '--update']])  # Restart with previous flags except for update flag.

try:  # Make sure all required libraries are installed. Appropriate requirements.txt is required. If a library which isn't included in any of these paths is the only one missing, just add an import to it anyways.
  from lxml import etree
  from communication_module import CommunicationModule
  import requests
  from py_logging.cestel_logging import init_logger, configure_default_logger, log, log_exception
except Exception as e:
  print('Some libraries are missing. Try running with --update flag:{}\n'.format(e))
  time.sleep(5)
  sys.exit(1)

if os.path.dirname(sys.argv[0]) != '':
  os.chdir(os.path.dirname(sys.argv[0]))
configure_default_logger(level=LOG_LEVEL, console=LOG_LEVEL)

# This is the logger that handles siwim_i file. It's one of two loggers (the other being update) that log both to file and console.
if sys.version < '3.8':  # Python2/3 use different formatting styles
  logger = init_logger('siwim_i', fmt='%(name)-9s %(asctime)s %(levelname)-8s %(message)s', to_console=True)
else:
  logger = init_logger('siwim_i', fmt='%(name)-9s %(asctime)s %(levelname)-8s [%(filename)s: %(lineno)d] %(message)s', to_console=True)

interrupted = False


fd = None
modules = None
module_names = list()

end_i = False

update_host = ""
update_protocol = "https"
sites_path = ""

comm_module_name = ""

version = [4, 10, 34]
channel = "stable"

log(logger, 'Running in debug mode.', sev=logging.DEBUG)

try:
  if i_conf_path == "":
    i_conf_path = "conf/conf.xml"
  fd = open(i_conf_path)
  doc = etree.parse(fd)
  fd.close()
  modules = list(doc.iter("module"))
  update_host = doc.find("update_host").text
  try:
    update_protocol = doc.find("update_protocol").text
  except:
    pass
  sites_path = doc.find("sites_path").text
  try:
    channel = doc.find("update_channel").text
  except:
    pass
  log(logger, "SiWIM-I v{} starting. Using {} update channel.".format(".".join(str(x) for x in version), channel))
  aggr_cnt = 0
  for module in modules:
    module_name = module.find("name").text
    if module_name in module_names:
      raise Exception("All module names must be unique!")
    module_names.append(module_name)
except:
  log_exception(logger, 'Parsing conf file failed:')
  sys.exit(0)

sites_path = sites_path.replace("\\", "/")
if sites_path[len(sites_path) - 1] == "/":
  sites_path = sites_path[:-1]  
if sites_path == "" or not os.path.exists(sites_path):
  log(logger, "Sites directory doesn't exist. Exiting ...", sev=logging.CRITICAL)
  sys.exit(0)

root_path = sites_path[0:sites_path.find("/sites")]

running = dict()
downstream = dict()  

comm_started = False

site_name = swmCurSite(root_path + "/siwim_e/conf/siwim.conf")

# factory
for module in modules:
  params = parse_module(module)
  params["root_path"] = root_path
  name = params["name"]
  if module.attrib["type"] == "rcv":
    from receive_module import ReceiveModule
    running[name] = ReceiveModule(params)
  elif module.attrib["type"] == "aggr":
    from aggregation_module import AggregationModule
    running[name] = AggregationModule(params)
  elif module.attrib["type"] == "cam": 
    from camera_module import CameraModule
    running[name] = CameraModule(params)
  elif module.attrib["type"] == "pic":  
    from picture_module import PictureModule
    running[name] = PictureModule(params)
  elif module.attrib["type"] == "out":
    from output_module import OutputModule
    running[name] = OutputModule(params)
  elif module.attrib["type"] == "lrm":
    from alarm_module import AlarmModule
    running[name] = AlarmModule(params)
  elif module.attrib["type"] == "stat":
    params["update_channel"] = channel
    params["version"] = version
    from status_module import StatusModule
    running[name] = StatusModule(params)
  elif module.attrib["type"] == "ctu":
    from ctu_module import CtuModule
    running[name] = CtuModule(params)
  elif module.attrib["type"] == "can":
    from can_module import CanModule
    running[name] = CanModule(params)
  elif module.attrib["type"] == "hist":
    from hist_module import HistModule
    running[name] = HistModule(params)
  elif module.attrib["type"] == "comm":
    comm_module_name = name
    running[name] = CommunicationModule(params)
    comm_started = True
  elif module.attrib["type"] == "papago":
    from papago_module import PapagoModule
    running[name] = PapagoModule(params)
  elif module.attrib["type"] == "efoy":
    from efoy_module import EfoyModule
    running[name] = EfoyModule(params)
  # add downstreams the old way
  try:
    running[name].set_sites_path(sites_path)
    running[name].set_site_name(site_name)
    dm_list = list(module.iter("downstream_module"))
    downstream[name] = list()
    for dm in dm_list:
      downstream[name].append(dm.text)
  except:
    log_exception(logger, 'Got exception retrieving downstream_module:')
  
if not comm_started:
  name = "comm_default"
  comm_module_name = name
  downstream[name] = list()
  running[name] = CommunicationModule({"name":name, "root_path":root_path})
      
for key, module in running.items(): 
  log(logger, "Starting " + key + " module ...")
  for downster in downstream[key]:
    try:
      module.add_downstream_module(running[downster])
    except:
      pass   
  # register alarms HERE (moved from receive module's run)
  for ds_ky, ds_mod in module.downstream_modules_dict.items():
    try:
      ds_mod.register_alarm(module.name, module.typ)
    except:
      pass
  threading.Thread(target=module.run).start()

# Whenever Ctrl+C is pressed, sig_handler, which sets interrupted variable, will be called.
signal.signal(signal.SIGINT, keyboard_interrupt)
day = 0
if print_memory_usage:  # The flag is set by passing a parameter
  try:
    from guppy import hpy  # This library causes a crash if used with win32com.client (which we don't use here)
    print_count = 60
  except:
    log(logger, 'Failed to import guppy. Please install guppy (Python2) or guppy3 (Python3) to enable -m flag.', sev=logging.WARNING)
    print_memory_usage = False
while True:
  try:
    if end_i:
      log(logger, 'Shutting down.')
      for key1, module1 in running.items():
        module1.set_end()
      while True:
        all_dead = True
        for key1, module1 in running.items():
          if module1.is_alive():
            log(logger, 'Waiting for {0} to die'.format(key1), sev=logging.DEBUG)
            all_dead = False
            break
        if all_dead:
          log(logger, 'All threads have closed.')
          break
        try:
          time.sleep(0.5)
        except:
          log(logger, 'Sleep failed?')
      log(logger, 'Calling os._exit(0) on siwim_i.py', sev=logging.DEBUG)
      #logLifeEvents(site_name, "i_shutdowns")
      os._exit(0)
    if print_memory_usage:
      if print_count == 60:
        print(hpy().heap())
        print_count = 0
      print_count += 1
    for key, module in running.items():
      if not module.is_alive():
        try:
          log(logger, "Restarting " + key + " module ...")
          threading.Thread(target=module.run).start()
        except:
          log_exception(logger, 'Got exception while trying to restart a module:')
    loc = ""
    new_ver_list = [0, 0, 0]
    try:
      if day != datetime.datetime.now().day and update_host != "":
        log(logger, 'Host: {0}, Channel: {1}'.format(update_host, channel), sev=logging.DEBUG)
        day = datetime.datetime.now().day
        resp = requests.post('{0}://{1}/check.php?sw=siwim-i&ch={2}'.format(update_protocol, update_host, channel), timeout=10).text
        resp_parts = resp.split()
        for part in resp_parts:
          if part[0:4] == "ver=":
            new_ver_list = part[4:].split(".")
            for i in range(len(new_ver_list)):
              new_ver_list[i] = int(new_ver_list[i])
          if part[0:4] == "loc=":
            loc = part[4:]
            log(logger, 'Update loc: {0}'.format(loc), sev=logging.DEBUG)
        loc = loc[loc.find("/"):]
        for i in range(len(version)):
          if version[i] < new_ver_list[i]:
            end_i = True
            break
          elif version[i] > new_ver_list[i]:
            break
    except:
      log(logger, 'No update found.', sev=logging.DEBUG)
    if end_i:
      log(logger, "Updating SiWIM-I to " + ".".join(str(x) for x in new_ver_list))
      if not os.path.exists("update"):
        os.mkdir("update/")
      f = requests.get("{0}://{1}{2}".format(update_protocol, update_host, loc), stream=True)
      with open('update/update.zip', 'wb') as fd:
        for data in f.iter_content(chunk_size=1024):
          fd.write(data)
      #delete files to make sure update directory is ready for use
      up_files = os.listdir("update")
      for up_file in up_files:
        if up_file != "update.zip" and up_file != "i_updater.py":
          os.remove("update/" + up_file)
      zip_ref = zipfile.ZipFile("update/update.zip", 'r')
      zip_ref.extractall("update")
      zip_ref.close()
      os.remove("update/update.zip")                               
      subprocess.Popen([sys.executable, "update/i_updater.py", str(os.getpid())])
    elif running[comm_module_name].are_we_resetting_1():
      log(logger, "Resetting I due to a frontend message ...")
      end_i = True
      if sys.platform == "win32":
        subprocess.Popen([sys.executable, "update/i_updater.py", str(os.getpid())])
    elif interrupted:
      log(logger, 'Received signal interrupt.')
      end_i = True
    else:
      try:
        new_name = swmCurSite()
        # if name changed, reset i
        # generic indicates that something went wrong with site read
        # if name was generic from the get go, okay
        if site_name != new_name and new_name != "generic":
          end_i = True
      except:
        pass
      try:
        time.sleep(1)
      except:
        log(logger, 'Exception during sleep (keyboard interrupt, maybe)')
  except:
    log_exception(logger, 'Runner, general exception:')
    
  